@@ -10,13 +10,14 @@ from box import Box
 
 # Related modules
 from .. import common
+from ..data import get_from_box
 from ..callback import Callback
 from ..augment.links import IFATTR
 from ..augment import devices
 
 # List of attributes we don't want propagated from defaults to global/node
 #
-no_propagate_list = ["attributes","extra_attributes","requires","supported_on","no_propagate"]
+no_propagate_list = ["attributes","extra_attributes","requires","supported_on","no_propagate","config_after","transform_after"]
 
 class _Module(Callback):
 
@@ -50,6 +51,24 @@ def pre_transform(topology: Box) -> None:
   node_transform("pre_transform",topology)
   link_transform("pre_transform",topology)
 
+"""
+pre_node_transform: executed just before the node data model transformation is started
+
+"""
+def pre_node_transform(topology: Box) -> None:
+  module_transform("pre_node_transform",topology)
+  node_transform("pre_node_transform",topology)
+  link_transform("pre_node_transform",topology)
+
+"""
+pre_link_transform: executed just before the link data model transformation is started
+
+"""
+def pre_link_transform(topology: Box) -> None:
+  module_transform("pre_link_transform",topology)
+  node_transform("pre_link_transform",topology)
+  link_transform("pre_link_transform",topology)
+
 """
 post_transform:
   execute module-specific code after the main link- and node
@@ -57,6 +76,7 @@ post_transform:
 """
 def post_transform(topology: Box) -> None:
   check_supported_node_devices(topology)       # A bit late, but we can do this check only after node data has been adjusted
+  common.exit_on_error()
   copy_node_data_into_interfaces(topology)     # Copy node attributes that match interface attributes into interfaces
   module_transform("post_transform",topology)
   node_transform("post_transform",topology)
@@ -105,6 +125,8 @@ def merge_node_module_params(topology: Box) -> None:
   for name,n in topology.nodes.items():
     if 'module' in n:
       for m in n.module:
+        if not m in topology.defaults:      # Cannot propagate parameters of unknown module
+          continue                          # ... error will be reported later
         if m in topology:
           global_settings = get_propagated_global_module_params(m,topology.get(m,{}),topology.defaults[m])
           if global_settings:
@@ -168,6 +190,10 @@ def adjust_global_modules(topology: Box) -> None:
   topology.module = list(mod_dict.keys())
   for m in topology.module:                                     # Iterate over all configured modules
     if not m in topology.defaults:                              # Does this module have defaults?
+      common.error(
+        f"Unknown module {m} (we found no system defaults for it)",
+        common.IncorrectValue,
+        'module')
       continue                                                  # Nope. Weird, but doesn't matter right now.
     mod_def = topology.defaults[m]
     if not isinstance(mod_def,dict):                               # Are module defaults a dict?
@@ -182,6 +208,22 @@ def adjust_global_modules(topology: Box) -> None:
     if len(default_copy):                        # Anything left? Let's merge it with existing settings
       topology[m] = default_copy + topology[m]   # Have to use this convoluted way to prevent generating empty dict
 
+  reorder_node_modules(topology,'transform_after')
+
+'''
+add_module_extra_parameters: add extra module keywords (ex: 'vrfs' for 'vrf' module) to the list of attributes
+'''
+def add_module_extra_parameters(topology: Box) -> None:
+  if not 'module' in topology:
+    return
+
+  for m in topology.module:                                     # Iterate through the global list of modules
+    if 'extra' in topology.defaults[m].attributes:              # Does the module have 'extra' parameters?
+      for k in topology.defaults[m].attributes.extra.keys():    # ... oh, it does, iterate through its keys (attribute levels)
+        for attr in topology.defaults[m].attributes.extra[k]:   # Take every attribute from the list of extra attributes
+          if not attr in topology.defaults.attributes[k]:       # ... and if it's not already in the global list of attributes
+            topology.defaults.attributes[k].append(attr)        # ... append it to the global list
+
 '''
 adjust_modules: somewhat intricate multi-step config module adjustments
 
@@ -194,6 +236,8 @@ def adjust_modules(topology: Box) -> None:
   augment_node_module(topology)
   merge_node_module_params(topology)
   adjust_global_modules(topology)
+  add_module_extra_parameters(topology)
+  module_transform("init",topology)
   check_module_parameters(topology)
   check_module_dependencies(topology)
   module_transform("pre_default",topology)
@@ -262,25 +306,45 @@ def check_module_parameters(topology: Box) -> None:
     for m in topology.get("module",[]):     # Iterate over all known modules - get avoids instantiating module key
       if m in l and mod_attr[m] and l[m]:   # ... focusing on modules that have attributes specified on this link
                                             # ... and want to have their attributes checked
-        for k in l[m].keys():               # Iterate over link-level module attributes
-          if not k in mod_attr[m].link:     # If the name of an attribute is not in the list of allowed
+
+        # Module attribute could be a list of keys or a class type
+        #
+        if isinstance(l[m],Box) and isinstance(mod_attr[m].link,list):
+          for k in l[m].keys():             # Iterate over link-level module attributes
+            if not k in mod_attr[m].link:   # If the name of an attribute is not in the list of allowed
                                             # ... link-level attributes, report an error
+              common.error(
+                f"Invalid attribute {k} for module {m} on link {l}",
+                common.IncorrectValue,
+                'modules')
+        else:
+          if type(l[m]).__name__ != str(mod_attr[m].link):
             common.error(
-              f"Invalid attribute {k} for module {m} on link {l}",
+              f"Invalid value type for attribute {k} for module {m} on link {l}, expected {mod_attr[m].link}",
               common.IncorrectValue,
               'modules')
 
-    for intf in l.interfaces:                   # Iterate over all interfaces attached to the link
-      n = intf.node                             # ... get node name
-      node = topology.nodes[n]                  # ... and node data structure (so the expressions don't get too crazy)
-      for m in node.get("module",[]):           # Iterate over all node modules
-        if mod_attr[m] and m in intf:           # Does the current module have a list of attributes?
-                                                # ... and interface contains module attributes?
-          for k in intf[m].keys():              # Iterate over node link-level module-specific attributes
-            if not k in mod_attr[m].interface:  # If the name of an attribute is not in the list of allowed
-                                                # ... interface attributes report error
+
+    for intf in l.interfaces:                     # Iterate over all interfaces attached to the link
+      n = intf.node                               # ... get node name
+      node = topology.nodes[n]                    # ... and node data structure (so the expressions don't get too crazy)
+      for m in node.get("module",[]):             # Iterate over all node modules
+        if mod_attr[m] and m in intf:             # Does the current module have a list of attributes?
+                                                  # ... and interface contains module attributes?
+          # Module attribute could be a list of keys or a class type
+          #
+          if isinstance(intf[m],Box) and isinstance(mod_attr[m].link,list):
+            for k in intf[m].keys():              # Iterate over node link-level module-specific attributes
+              if not k in mod_attr[m].interface:  # If the name of an attribute is not in the list of allowed
+                                                  # ... interface attributes report error
+                common.error(
+                  f"Node {n} has invalid attribute {k} for module {m} on link {l}",
+                  common.IncorrectValue,
+                  'modules')
+          else:
+            if type(intf[m]).__name__ != str(mod_attr[m].interface):
               common.error(
-                f"Node {n} has invalid attribute {k} for module {m} on link {l}",
+                f"Node {n} has invalid value type for attribute {k} for module {m} on link {l}, expected {mod_attr[m].interface}",
                 common.IncorrectValue,
                 'modules')
 
@@ -298,7 +362,8 @@ Otherwise:
 def parse_module_attributes(a: typing.Union[typing.Dict, Box]) -> Box:
   if isinstance(a,dict):
     attr = Box(a,default_box=True,box_dots=True)
-    attr.interface = list(set(attr.link) - set(attr.link_no_propagate) | set(attr.interface))
+    if not isinstance(attr.interface,str):
+      attr.interface = list(set(attr.link) - set(attr.link_no_propagate) | set(attr.interface))
     attr.node = attr.get("node",attr["global"])
   else:
     attr = Box({
@@ -327,16 +392,20 @@ def check_module_dependencies(topology:  Box) -> None:
 """
 reorder_node_modules:
 
-For every node with at least two modules: sort the list of modules based on their dependencies preserving
-the original order as much as possible.
+For every node with at least two modules: sort the list of modules based on their _requires_ and
+_config_after_/_transform_after_ dependencies preserving the original order as much as possible.
 """
 
-def reorder_node_modules(topology: Box) -> None:
+def reorder_node_modules(topology: Box, secondary_sort: str = "config_after") -> None:
+  if 'module' in topology:
+    topology.module = sort_module_list(topology.module,topology.defaults, secondary_sort)
+    topology.defaults.module = topology.module
+
   for name,n in topology.nodes.items():
     if 'module' in n:
-      n.module = sort_module_list(n.module,topology.defaults)
+      n.module = sort_module_list(n.module,topology.defaults, secondary_sort)
 
-def sort_module_list(mods: list, mod_params: Box) -> list:
+def sort_module_list(mods: list, mod_params: Box, secondary_sort: str = "config_after") -> list:
   if (len(mods) < 2):
     return mods
 
@@ -345,7 +414,7 @@ def sort_module_list(mods: list, mod_params: Box) -> list:
     skipped: typing.List[str] = []
     for m in mods:
       if m in mod_params:
-        requires = mod_params[m].get('requires',[])
+        requires = mod_params[m].get('requires',[]) + mod_params[m].get(secondary_sort,[])
         if [ r for r in requires if r in mods ]:
           skipped = skipped + [ m ]
         else:
@@ -376,6 +445,37 @@ def copy_node_data_into_interfaces(topology: Box) -> None:
             if isinstance(intf.get(m,{}),dict):                  # .. but only if the interface data allows it
               intf[m] = copy_attr + intf[m]
 
+"""
+get_effective_module_attribute:
+
+Walk through the inheritance chain as supplied by the caller and extract desired module attribute
+"""
+def get_effective_module_attribute(
+    path: str,
+    intf: typing.Optional[Box] = None,
+    link: typing.Optional[Box] = None,
+    node: typing.Optional[Box] = None,
+    topology: typing.Optional[Box] = None,
+    defaults: typing.Optional[Box] = None,
+    merge_dict: bool = True) -> typing.Optional[typing.Any]:
+
+  composite_value: typing.Optional[Box] = None
+  for obj in (intf,link,node,topology,defaults):
+    if obj is None:
+      continue
+    value = get_from_box(obj,path)
+    if not value:
+      continue
+    if not isinstance(value,Box) or not merge_dict:
+      return value
+    if composite_value:
+      composite_value = value + composite_value
+    else:
+      composite_value = value
+
+  return composite_value
+
+
 """
 Callback transformation routines
 
@@ -393,6 +493,9 @@ def module_transform(method: str, topology: Box) -> None:
   for m in topology.get('module',[]):
     if not mod_load.get(m):
       mod_load[m] = _Module.load(m,topology.get(m))
+    if common.DEBUG:
+      if hasattr(mod_load[m],f"module_{method}"):
+        print(f'Calling module {m} module_{method}')
     mod_load[m].call("module_"+method,topology)
 
 def node_transform(method: str , topology: Box) -> None:
@@ -402,6 +505,9 @@ def node_transform(method: str , topology: Box) -> None:
     for m in n.get('module',[]):
       if not mod_load.get(m):  # pragma: no cover (module should have been loaded already)
         mod_load[m] = _Module.load(m,topology.get(m))
+      if common.DEBUG:
+        if hasattr(mod_load[m],f"node_{method}"):
+          print(f'Calling module {m} node_{method} on node {name}')
       mod_load[m].call("node_"+method,n,topology)
 
 def link_transform(method: str, topology: Box) -> None:
@@ -414,4 +520,7 @@ def link_transform(method: str, topology: Box) -> None:
     for m in mod_list.keys():
       if not mod_load.get(m):  # pragma: no cover (module should have been loaded already)
         mod_load[m] = _Module.load(m,topology.get(m))
+      if common.DEBUG:
+        if hasattr(mod_load[m],f"link_{method}"):
+          print(f'Calling module {m} link_{method} on link {l.name}')
       mod_load[m].call("link_"+method,l,topology)
