@@ -9,14 +9,17 @@ import netaddr
 
 from . import _Module,_routing
 from .. import common
+from .. import data
 from ..augment.links import IFATTR
 
 def check_bgp_parameters(node: Box) -> None:
   if not "bgp" in node:  # pragma: no cover (should have been tested and reported by the caller)
     return
-  if not "as" in node.bgp or not isinstance(node.bgp.get('as',{}),int):
+  if not "as" in node.bgp:
     common.error("Node %s has BGP enabled but no AS number specified" % node.name)
 
+  data.must_be_int(parent=node,key='bgp.as',path=f'nodes.{node.name}',min_value=1,max_value=65535,module='bgp')
+
   if "community" in node.bgp:
     bgp_comm = node.bgp.community
     if isinstance(bgp_comm,str):
@@ -24,17 +27,22 @@ def check_bgp_parameters(node: Box) -> None:
     elif isinstance(bgp_comm,list):
       node.bgp.community = { 'ibgp' : bgp_comm, 'ebgp': bgp_comm }
     elif not(isinstance(bgp_comm,dict)):
-      common.error("bgp.community attribute in node %s should be a string, a list, or a dictionary (%s)" % (node.name,str(bgp_comm)))
+      common.error(
+        f"bgp.community attribute in node {node.name} should be a string, a list, or a dictionary (found: {bgp_comm})",
+        common.IncorrectType,
+        'bgp')
       return
 
     for k in node.bgp.community.keys():
       if not k in ['ibgp','ebgp']:
         common.error("Invalid BGP community setting in node %s: %s" % (node.name,k))
-      if isinstance(node.bgp.community[k],str):
-        node.bgp.community[k] = [ node.bgp.community[k] ]
-      for v in node.bgp.community[k]:
-        if not v in ['standard','extended']:
-          common.error("Invalid BGP community propagation setting for %s sessions in node %s: %s" % (k,node.name,v))
+      else:
+        data.must_be_list(
+          parent=node.bgp.community,
+          path=f'nodes.{node.name}.bgp.community',
+          key=k,
+          valid_values=['standard','extended'],
+          module='bgp')
 
 def find_bgp_rr(bgp_as: int, topology: Box) -> typing.List[Box]:
   return [ n 
@@ -81,23 +89,23 @@ class BGP(_Module):
       return
 
     node_data = Box({},default_box=True,box_dots=True)
-    for asn,data in topology.bgp.as_list.items():
-      if not isinstance(data,Box):
+    for asn,as_data in topology.bgp.as_list.items():
+      if not isinstance(as_data,Box):
         common.error(
           "Invalid value in bgp.as_list for ASN %s: " % asn + \
           "\n... Each ASN in a BGP as_list must be a dictionary with (at least) members key:"+
-          "\n... Found: %s" % data,
+          "\n... Found: %s" % as_data,
           common.IncorrectValue)
         continue
 
-      if not 'members' in data:
+      if not 'members' in as_data:
         common.error(
           "BGP as_list for ASN %s does not have a member attribute" % asn,
           common.IncorrectValue)
         continue
 
-      common.must_be_list(data,'members',f'bgp.as_list.{asn}')
-      for n in data.members:
+      data.must_be_list(as_data,'members',f'bgp.as_list.{asn}')
+      for n in as_data.members:
         if not n in topology.nodes:
           common.error(
             "Invalid node name %s in member list of BGP AS %s" % (n,asn),
@@ -110,7 +118,7 @@ class BGP(_Module):
           continue
         node_data[n]["as"] = asn
 
-      for n in data.get('rr',{}):
+      for n in as_data.get('rr',{}):
         if not n in topology.nodes:
           common.error(
             "Invalid node name %s in route reflector list of BGP AS %s" % (n,asn),
@@ -239,7 +247,14 @@ class BGP(_Module):
           if "unnumbered" in l:
             extra_data.unnumbered = True
             extra_data.local_if = l.ifname
-          node.bgp.neighbors.append(bgp_neighbor(neighbor,ngb_ifdata,'ebgp',extra_data))
+
+          ebgp_data = bgp_neighbor(neighbor,ngb_ifdata,'ebgp',extra_data)
+          if 'vrf' in l:        # VRF neighbor
+            if not node.vrfs[l.vrf].bgp.neighbors:
+              node.vrfs[l.vrf].bgp.neighbors = []
+            node.vrfs[l.vrf].bgp.neighbors.append(ebgp_data)
+          else:                 # Global neighbor
+            node.bgp.neighbors.append(ebgp_data)
 
     # Calculate BGP address families
     #
