@@ -79,6 +79,43 @@ def get_next_vlan_id(k : str) -> int:
   vlan_ids[k].add(vlan_next[k])
   return vlan_next[k]
 
+#
+# routed_access_vlan: Given a link with access/native VLAN, check if all nodes on the link use routed VLAN
+#
+def routed_access_vlan(link: Box, topology: Box, vlan: str) -> bool:
+  def_link  = get_from_box(link,'vlan.mode')
+  def_vlan  = get_from_box(topology,f'vlans.{vlan}.mode')
+  def_global = get_from_box(topology,'vlan.mode') or 'irb'
+
+  #print(f'RAV: {link}')
+  #print(f'RAV: vlan {vlan} def_mode {def_mode}')
+  for intf in link.interfaces:
+    mode = get_from_box(intf,'vlan.mode') or \
+           def_link or \
+           get_from_box(topology.nodes[intf.node],f'vlans.{vlan}.mode') or \
+           get_from_box(topology.nodes[intf.node],'vlan.mode') or \
+           def_vlan or \
+           def_global or 'irb'
+    if mode != 'route':
+      return False
+
+  #print(f'RAV: routed VLAN')
+  return True
+
+#
+# interface_vlan_mode: Given an interface, a node, and topology, find interface VLAN mode
+#
+def interface_vlan_mode(intf: Box, node: Box, topology: Box) -> str:
+  vlan = get_from_box(intf,'vlan.access') or get_from_box(intf,'vlan.native')
+  if not vlan:
+    return 'irb'
+
+  return get_from_box(intf,'vlan.mode') or \
+         get_from_box(node,f'vlans.{vlan}.mode') or \
+         get_from_box(node,'vlan.mode') or \
+         get_from_box(topology,f'vlans.{vlan}.mode') or \
+         get_from_box(topology,'vlan.mode') or 'irb'
+
 #
 # Validate VLAN attributes and set missing attributes:
 #
@@ -108,7 +145,7 @@ def validate_vlan_attributes(obj: Box, topology: Box) -> None:
     vdata = obj.vlans[vname]
 
     if 'mode' in vdata:                                             # Do we have 'mode' set in the VLAN definition?
-      if not vdata.mode in vlan_mode_kwd:                           # ... check the keyword value 
+      if not vdata.mode in vlan_mode_kwd:                           # ... check the keyword value
         common.error(
           f'Invalid VLAN mode setting {vdata.mode} in VLAN {vname} in {obj_name}',
           common.IncorrectValue,
@@ -117,8 +154,6 @@ def validate_vlan_attributes(obj: Box, topology: Box) -> None:
       if default_fwd_mode:                                          # Propagate default VLAN forwarding mode if needed
         vdata.mode = default_fwd_mode
 
-    if 'mode' in vdata and vdata.mode == 'route':                   # Throw out attempts to have routed interfaces
-      common.error(f'VLAN routed interfaces not yet supported: VLAN {vname} in {obj_name}',common.IncorrectValue,'vlan')
     if not 'id' in vdata:                                           # When VLAN ID is not defined
       vdata.id = get_next_vlan_id('id')                             # ... take the next free VLAN ID from the list
     if not isinstance(vdata.id,int):                                # Now validate the heck out of VLAN ID
@@ -190,7 +225,7 @@ def check_link_vlan_attributes(obj: Box, link: Box, v_attr: Box, topology: Box)
         common.error(
           f'VLAN attribute {attr}{node_error} must be a {a_type.__name__}\n... {link}',
           common.IncorrectValue,
-          'vlan')  
+          'vlan')
         link_ok = False
         continue
 
@@ -284,7 +319,7 @@ def validate_link_vlan_attributes(link: Box,v_attr: Box,topology: Box) -> bool:
         f'Native VLAN {v_attr.native.list[0]} used on a link is not in the VLAN trunk definition\n... {link}',
         common.IncorrectValue,
         'vlan')
-    else:               
+    else:
       for intf in link.interfaces:                                # Now check if every node using native VLAN has it in its trunk
         if 'vlan' in intf:                                        # VLAN attributes on interface?
                                                                   # Calculate effective node trunk and native VLAN data
@@ -323,6 +358,29 @@ def copy_vlan_attributes(vlan: str, vlan_data: Box, link: Box) -> None:
     link.type = vlan_data.get('type','lan')
   link.vlan_name = vlan
 
+"""
+get_link_access_vlan: given v_attr data structure, find usable access or native vlan
+"""
+def get_link_access_vlan(v_attr: Box) -> typing.Optional[str]:
+  if 'access' in v_attr:
+    if v_attr.access.set:
+      return list(v_attr.access.set)[0]
+
+  if 'native' in v_attr:
+    if v_attr.native.set:
+      return list(v_attr.native.set)[0]
+
+  return None
+
+"""
+fix_vlan_mode_attribute: Given an interface/link data structure, replace 'mode' with 'vlan.mode'
+"""
+def fix_vlan_mode_attribute(data: Box) -> None:
+  if 'mode' in data:
+    data.vlan.mode = data.mode
+    data.pop('mode',None)
+
+
 """
 set_link_vlan_prefix: copy link attributes from VLAN for access/native VLAN links
 
@@ -374,6 +432,7 @@ def create_vlan_links(link: Box, v_attr: Box, topology: Box) -> None:
       link_data.vlan_name = vname
       link_data.type = 'vlan_member'
       link_data.interfaces = []
+      fix_vlan_mode_attribute(link_data)
 
       if vname in topology.get('vlans',{}):                 # We need an IP prefix for the VLAN link
         prefix = topology.vlans[vname].prefix               # Hopefully we can get it from the global VLAN pool
@@ -382,12 +441,20 @@ def create_vlan_links(link: Box, v_attr: Box, topology: Box) -> None:
         if 'vlan' in intf and vname in intf.vlan.get('trunk',{}):
           intf_data = Box(intf.vlan.trunk[vname] or {},default_box=True,box_dots=True)
           intf_data.node = intf.node
+          if 'mode' in intf.vlan and not get_from_box(intf_data,'vlan.mode'):
+            intf_data.vlan.mode = intf.vlan.mode            # vlan.mode is inherited from trunk dictionary or parent interface
           link_data.interfaces.append(intf_data)
           if not prefix:                                    # Still no usable IP prefix? Try to get it from the node VLAN pool
             if vname in topology.nodes[intf.node].get('vlans',{}):
               prefix = topology.node[intf.node].vlans[vname].prefix
 
-      link_data.prefix = prefix
+      if routed_access_vlan(link_data,topology,vname):
+        link_data.vlan.mode = 'route'
+        for intf in link_data.interfaces:
+          intf.vlan.mode = 'route'
+      else:
+        link_data.prefix = prefix
+
       topology.links.append(link_data)
 
 """
@@ -431,7 +498,7 @@ create_node_vlan: Create a local (node) copy of a VLAN used on an interface
 def create_node_vlan(node: Box, vlan: str, topology: Box) -> typing.Optional[Box]:
   if not vlan in node.vlans:                                        # Do we have VLAN defined in the node?
     node.vlans[vlan] = Box(topology.vlans[vlan])                    # ... no, create a copy of the global definition
-    if not node.vlans[vlan]:                                        # pragma: no cover -- we don't have a global definition? 
+    if not node.vlans[vlan]:                                        # pragma: no cover -- we don't have a global definition?
       common.fatal(                                                 # ... this should have been detected way earlier
         f'Unknown VLAN {vlan} used on node {node.name}','vlan')
       return None
@@ -464,10 +531,32 @@ def create_svi_interfaces(node: Box, topology: Box) -> dict:
     if not access_vlan:                                                     # No access VLAN on this interface?
       continue                                                              # ... good, move on
 
+    routed_intf = interface_vlan_mode(ifdata,node,topology) == 'route'      # Identify routed VLAN interfaces
+    vlan_subif  = routed_intf and ifdata.get('type','') == 'vlan_member'    # ... and VLAN-based subinterfaces
+
     vlan_data = create_node_vlan(node,access_vlan,topology)
-    if vlan_data is None:
+    if vlan_data is None:                                                   # pragma: no-cover
+      if vlan_subif:                                                        # We should never get here, but at least we can
+        common.fatal(                                                       # scream before crashing
+          f'Weird: cannot get VLAN data for VLAN {access_vlan} on node {node.name}, aborting')
       continue
 
+    if routed_intf:                                                         # Routed VLAN access interface, turn it back into native interface
+      for k in ('access','native'):                                         # ... remove access VLAN attributes
+        ifdata.vlan.pop(k,None)
+
+      if not ifdata.vlan:                                                   # ... and VLAN dictionary if there's nothing else left
+        ifdata.pop('vlan',None)
+
+      vlan_copy = { k:v for (k,v) in vlan_data.items() if not k in svi_skipattr and k != 'mode' }
+
+      if vlan_subif:
+        ifdata.vlan.mode = 'route'
+        ifdata.vlan.access_id = vlan_data.id
+
+      node.interfaces[ifidx] = vlan_copy + ifdata                           # Merge VLAN data with interface data
+      continue                                                              # Move to next interface
+
     features = devices.get_device_features(node,topology.defaults)
     svi_name = features.vlan.svi_interface_name
     if not svi_name:                                                        # pragma: no cover -- hope we got device settings right ;)
@@ -479,22 +568,27 @@ def create_svi_interfaces(node: Box, topology: Box) -> dict:
 
     if not access_vlan in vlan_ifmap:                                       # Do we need to create a SVI interface?
       skip_attr = list(skip_ifattr)                                         # Create a local copy of the attribute skip list
-      if vlan_data.get('mode','') == 'bridge':                              # ... and skip IP addresses for bridging-only VLANs
+      vlan_mode = ifdata.vlan.get('mode','') or vlan_data.get('mode','')    # Get VLAN forwarding mode
+      if vlan_mode == 'bridge':                                             # ... and skip IP addresses for bridging-only VLANs
         skip_attr.extend(['ipv4','ipv6'])
       vlan_ifdata = Box(                                                    # Copy non-physical interface attributes into SVI interface
         { k:v for k,v in ifdata.items() if k not in skip_attr },            # ... that will also give us IP addresses
         default_box=True,
         box_dots=True)
+      if vlan_mode:                                                         # Set VLAN forwarding mode for completness' sake
+        vlan_ifdata.vlan.mode = vlan_mode
       vlan_ifdata.ifindex = node.interfaces[-1].ifindex + 1                 # Fill in the rest of interface data:
       vlan_ifdata.ifname = svi_name.format(                                 # ... ifindex, ifname, description
                               vlan=vlan_data.id,
-                              bvi=vlan_data.bridge_group)
+                              bvi=vlan_data.bridge_group,
+                              ifname=ifdata.ifname)
       vlan_ifdata.name = f'VLAN {access_vlan} ({vlan_data.id})'
       vlan_ifdata.virtual_interface = True                                  # Mark interface as virtual
       vlan_ifdata.type = "svi"
       vlan_ifdata.neighbors = []                                            # No neighbors so far
                                                                             # Overwrite interface settings with VLAN settings
       vlan_ifdata = vlan_ifdata + { k:v for k,v in vlan_data.items() if k not in svi_skipattr }
+      fix_vlan_mode_attribute(vlan_ifdata)
       node.interfaces.append(vlan_ifdata)                                   # ... and add SVI interface to list of node interfaces
       vlan_ifmap[access_vlan] = vlan_ifdata
     else:
@@ -581,16 +675,31 @@ def rename_vlan_subinterfaces(node: Box, topology: Box) -> None:
   skip_ifattr.extend(topology.defaults.providers.keys())
 
   features = devices.get_device_features(node,topology.defaults)
-  subif_name = features.vlan.vlan_subif_name
+  if 'switch' in features.vlan.model:                             # No need for VLAN subinterfaces, remove non-routed vlan_member interfaces
+    node.interfaces = [ intf for intf in node.interfaces \
+                          if intf.type != 'vlan_member' or intf.vlan.get('mode','') == 'route' ]
+    subif_name = features.vlan.routed_subif_name                  # Just in case: try to get interface name pattern for routed subinterface
 
-  if not subif_name:
-    node.interfaces = [ intf for intf in node.interfaces if intf.type != 'vlan_member']
-    return
-
-  for intf in node.interfaces:
+  for intf in node.interfaces:                                    # Now loop over remaining vlan_member interfaces
     if intf.type != 'vlan_member':
       continue
 
+    if not 'router' in features.vlan.model and not 'l3-switch' in features.vlan.model:
+      #
+      # The only way to get here is to have a routed VLAN in a VLAN trunk on a device that
+      # does not support VLAN subinterfaces
+      #
+      common.error(
+        f'Routed subinterfaces on VLAN trunks not supported on device type {node.device}\n... node {node.name} vlan {intf.vlan_name}',
+        common.IncorrectValue,'vlan')
+      continue
+
+    subif_name = features.vlan.subif_name
+    if not subif_name:                                            # pragma: no-cover -- hope we got device settings right
+      common.fatal(
+        f'Internal error: device {node.device} acts as a VLAN-capable {features.vlan.model} but does not have subinterface name template')
+      continue
+
     parent_intf = find_parent_interface(intf,node,topology)
     if parent_intf is None:
       common.fatal(f'Internal error: cannot find parent interface for {intf} in node {node.name}')
@@ -599,15 +708,77 @@ def rename_vlan_subinterfaces(node: Box, topology: Box) -> None:
     if 'subif_index' in parent_intf:
       parent_intf.subif_index = parent_intf.subif_index + 1
     else:
-      parent_intf.subif_index = 1
+      parent_intf.subif_index = features.vlan.first_subif_id or 1
 
-    intf.ifname = subif_name.format(**parent_intf)
+    ifname_data = parent_intf + intf                                  # Add parent interface data to subinterface data
+    ifname_data.ifname = parent_intf.ifname                           # ... making sure ifname is coming from parent interface
+
+    old_intf = Box({ 'ifname': intf.ifname })                         # Create a fake interface with old interface name
+    intf.ifname = subif_name.format(**ifname_data)
     intf.parent_ifindex = parent_intf.ifindex
+    intf.parent_ifname = parent_intf.ifname
     intf.virtual_interface = True
+    if 'vlan_name' in intf:
+      update_vlan_neighbor_list(intf.vlan_name,old_intf,intf,node,topology)
+
     for attr in skip_ifattr:
       if attr in intf and not attr in keep_subif_attr:
         intf.pop(attr,None)
 
+    if intf.vlan.access_id in parent_intf.vlan.get('trunk_id',[]):    # Remove subinterface VLAN from parent interface trunk list
+      parent_intf.vlan.trunk_id = [ vlan for vlan in parent_intf.vlan.trunk_id if vlan != intf.vlan.access_id ]
+      if not parent_intf.vlan.trunk_id:                               # Nothing left?
+        parent_intf.vlan.pop('trunk_id',None)                         # ... remove all trunk-related attributes from parent interface
+        parent_intf.vlan.pop('trunk',None)
+        if not parent_intf.vlan:                                      # And remove the VLAN dictionary if it's no longer needed
+          parent_intf.pop('vlan',None)
+
+  # Final check: mixed bridged/routed trunks
+  #
+  # We can skip this check if the device supports mixed routed/bridged trunks or if it
+  # uses VLAN subinterfaces to implement VLANs.
+  #
+  if features.vlan.mixed_trunk:
+    return
+
+  err_ifmap = {}
+  for intf in node.interfaces:
+    if not intf.get('parent_ifindex') or intf.type != 'vlan_member':  # Skip everything that is not a VLAN subinterface
+      continue
+
+    parent_intf_list = [ x for x in node.interfaces if x.ifindex == intf.parent_ifindex ]
+    if not parent_intf_list:
+      common.fatal(f'Internal error: cannot find parent interface for {intf} in node {node.name}')
+      return
+
+    parent_intf = parent_intf_list[0]
+    if parent_intf is None \
+         or get_from_box(parent_intf,'vlan.trunk_id') is None:        # No VLAN trunk left on the parent interface?
+      continue                                                        # ... cool, we're done
+
+    if not parent_intf.ifindex in err_ifmap:                          # We have a problem. Do we have to generate an error?
+      common.error(
+        f'Device type {node.device} does not support mixed bridged/routed VLAN trunks\n'+ \
+        f'... node {node.name} interface {parent_intf.ifname}: {parent_intf.name}',
+        common.IncorrectValue,
+        'vlan')
+    err_ifmap[parent_intf.ifindex] = True
+
+"""
+cleanup_vlan_name -- remove internal 'vlan_name' attribute from links and interfaces
+"""
+
+def cleanup_vlan_name(topology: Box) -> None:
+  if 'links' in topology:
+    for l in topology.links:
+      if 'vlan_name' in l:
+        l.pop('vlan_name',None)
+
+  for n in topology.nodes.values():
+    for intf in n.interfaces:
+      if 'vlan_name' in intf:
+        intf.pop('vlan_name',None)
+
 class VLAN(_Module):
 
   def module_pre_transform(self, topology: Box) -> None:
@@ -659,7 +830,26 @@ class VLAN(_Module):
     if not validate_link_vlan_attributes(link,v_attr,topology):
       return
 
-    set_link_vlan_prefix(link,v_attr,topology)
+    svi_skipattr = ['id','vni','mode','pool','prefix']                             # VLAN attributes not copied into link data
+    link_vlan = get_link_access_vlan(v_attr)
+    routed_vlan = False
+    if not link_vlan is None:
+      routed_vlan = routed_access_vlan(link,topology,link_vlan)
+      vlan_data = get_from_box(topology,f'vlans.{link_vlan}')                      # Get global VLAN data
+      if isinstance(vlan_data,Box):
+        vlan_data = Box({ k:v for (k,v) in vlan_data.items() \
+                                if k not in svi_skipattr })                       # Remove VLAN-specific data
+        fix_vlan_mode_attribute(vlan_data)                                        # ... and turn mode into vlan.mode
+        for (k,v) in vlan_data.items():                                           # Now add the rest to link data
+          if not k in link:                                                       # ... have to do the deep merge manually as
+            link[k] = v                                                           # ... we cannot just replace link data structure
+          elif isinstance(link[k],Box) and isinstance(vlan_data[k],Box):
+            link[k] = vlan_data[k] + link[k]
+
+    if routed_vlan:
+      link.vlan.mode = 'route'
+    else:
+      set_link_vlan_prefix(link,v_attr,topology)
 
     # Merge link VLAN attributes into interface VLAN attributes to make subsequent steps easier
     if 'vlan' in link:
@@ -681,3 +871,5 @@ class VLAN(_Module):
       set_svi_neighbor_list(n,topology)
 
     topology.links = [ link for link in topology.links if link.type != 'vlan_member' ]
+
+    cleanup_vlan_name(topology)
